import logging
from pathlib import Path
from typing import Union

import pandas as pd

_logger = logging.getLogger(__name__)


def parse_col_for_limits(columnname: str):
    """
    Return the upper and lower specification limits embedded into the column header.

    Examples:

      * ``speed`` - just a column header called "speed"
      * ``speed (lsl=10.4)`` - a column header with a lower specification limit built-in (no upper)
      * ``speed (lsl=10.4 usl=15.1)`` - a column header with both lower and upper specification limits

    :param columnname: the column name to parse
    :return: 2-value tuple of the form `(lsl, usl)`; returns `(None, None)` if values not found
    """
    lsl, usl = None, None
    if "(" in columnname and ")" in columnname:
        _logger.info(f"parentheses detected, loading thresholds")
        strings = columnname.split("(")[1].replace(")", "")
        strings = strings.strip()
        if not strings:
            return None, None

        parts = strings.split(" ")

        for part in parts:
            if "lsl" in part.lower():
                lsl_str = part.split("=")[1]
                try:
                    lsl = int(lsl_str)
                except ValueError:
                    lsl = float(lsl_str)
            elif "usl" in part.lower():
                usl_str = part.split("=")[1]
                try:
                    usl = int(usl_str)
                except ValueError:
                    usl = float(usl_str)

    return lsl, usl


def import_csv(file_path: (str, Path), columnname: str, **kwargs) -> Union[dict, pd.Series]:
    """
    Imports data from a csv file and outputs the specified column of data as a `pandas.Series`

    :param file_path: the path to the file on the local file system
    :param columnname: the column name to which the data is associated
    :param kwargs: keyword arguments to be passed directly into `pandas.read_csv()`
    :return: a dict containing a pandas series and the limits of the data to be analyzed
    """
    df = pd.read_csv(file_path, **kwargs)

    lsl, usl = parse_col_for_limits(columnname)

    if lsl is None and usl is None:
        return df[columnname]
    else:
        data = {"data": df[columnname]}

    if lsl is not None:
        data["lower_specification_limit"] = lsl
    if usl is not None:
        data["upper_specification_limit"] = usl

    return data


def import_excel(file_path: (str, Path), columnname: str, **kwargs) -> Union[dict, pd.Series]:
    """
    Imports data from an excel file and outputs the specified column of data as a `pandas.Series`

    :param file_path: the path to the file on the local file system
    :param columnname: the column name to which the data is associated
    :param kwargs: keyword arguments to be passed directly into `pandas.read_excel()`
    :return: a pandas series of the data which is to be analyzed
    """
    df = pd.read_excel(file_path, **kwargs)

    lsl, usl = parse_col_for_limits(columnname)

    if lsl is None and usl is None:
        return {"data": df[columnname]}
    else:
        data = {"data": df[columnname]}

    if lsl is not None:
        data["lower_specification_limit"] = lsl
    if usl is not None:
        data["upper_specification_limit"] = usl

    return data



import logging
from typing import List

import pandas as pd
import numpy as np

_logger = logging.getLogger(__name__)


def coerce(data: (List[int], List[float], pd.Series, np.array)) -> pd.Series:
    """
    Ensures that the data is of a type that can be easily manipulated. and eliminates extreme outliers.

    :param data: a list or list-like iterable
    :return: a pandas Series
    """
    if not isinstance(data, pd.Series):
        _logger.debug(f"attempting to convert data into pandas.Series...")
        data = pd.Series(data)
        _logger.debug(f"...conversion successful")

    if not isinstance(data, pd.Series):
        raise ValueError(
            "data is not of the correct type; expecting a list of integers, "
            "floats, a pandas.Series, or numpy.array"
        )

    return data


def remove_outliers(data: pd.Series) -> "pd.Series":
    # when data is considered an extreme outlier,
    # then we will re-scale the y limits
    data = data.copy().dropna()

    q25 = data.quantile(0.25)
    q50 = data.quantile(0.50)
    q75 = data.quantile(0.75)
    iqr = (q75 - q25) * 2
    min_data = q50 - (iqr * 1.5)
    max_data = q50 + (iqr * 1.5)

    data = data[(data > min_data) & (data < max_data)]

    return data.dropna().reset_index(drop=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['manufacturing']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0',
 'openpyxl>=3.0.9,<4.0.0',
 'pandas>=1.4.2,<2.0.0',
 'scipy>=1.7.0,<2.0.0',
 'xlrd>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'manufacturing',
    'version': '1.1.25',
    'description': 'Six-Sigma based analysis of manufacturing data',
    'long_description': '# Purpose\n\nTo provide analysis tools and metrics useful in manufacturing environments.\n\nGo to the [documentation](https://slightlynybbled.github.io/manufacturing/index.html).\n\n# Project Maturity\n\nProject is currently undergoing frequent updates for documentation and to add\nfunctionality and update documentation!!  Screenshots and features that you see\non here may be out of date, but are in progress.\n\nCurrent focus is to add test coverage and shore up functionality that has \nbeen implemented.  All versions before v1.2.0 will be minor bugfixes or minor \nfunctionality as it is revealed in test development.\n\n# Installation\n\nTo install from `pypi`:\n\n    pip install manufacturing\n\nTo install from source download and install using poetry:\n\n    poetry install\n\n# Usage\n\n## Cpk Visualization\n\nThe most useful feature of the `manufacturing` package is the visualization of Cpk.\nAs hinted previously, the `ppk_plot()` function is the primary method for display of\nCpk visual information.  First, get your data into a `list`, `numpy.array`, or \n`pandas.Series`; then supply that data, along with the `lower_control_limit` and \n`upper_control_limit` into the `ppk_plot()` function.\n\n    manufacturing.ppk_plot(data, lower_specification_limit=-2, upper_specification_limit=2)\n    \n![Screenshot](images/example3.png)\n\nIn this example, it appears that the manufacturing processes are not up to the task of \nmaking consistent product within the specified limits.\n\n## Zone Control Visualization\n\nAnother useful feature is the zone control visualization.\n\n    manufacturing.control_chart(data)\n\nThere are X-MR charts, Xbar-R charts, and Xbar-S charts available as well.  If you call the \n`control_chart()` function, the appropriate sample size will be selected and data grouped as\nthe dataset requires.  However, if you wish to call a specific type of control chart, use\n\n - `x_mr_chart`\n - `xbar_r_chart`\n - `xbar_s_chart`\n\n# Contributions\n\nContributions are welcome!  \n\n## RoadMap\n\nItems marked out were added most recently.\n\n - ...\n - ~~Add use github actions for deployment~~\n - ~~Transition to `poetry` for releases~~\n - ~~Add `I-MR Chart` (see `examples/imr_chart.py`)~~\n - ~~Add `Xbar-R Chart` (subgroups between 2 and 10)~~\n - ~~Add `Xbar-S Chart` (subgroups of 11 or more)~~\n - ~~Update documentation to reflect recent API changes~~\n - Add automated testing\n\n# Gallery\n\n![Ppk example](docs/_static/images/ppk_plot.png)\n\n![Cpk example](docs/_static/images/cpk_plot.png)\n\n![X-MR Chart](docs/_static/images/xmr_chart.png)\n\n![Xbar-R Chart](docs/_static/images/xbarr_chart.png)\n\n![Xbar-S Chart](docs/_static/images/xbars_chart.png)\n',
    'author': 'Jason R. Jones',
    'author_email': 'slightlynybbled@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import logging
from typing import List, Optional, Union

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axis import Axis
import numpy as np
import pandas as pd
import scipy.stats as stats

from manufacturing.analysis import (
    calc_ppk,
    control_beyond_limits,
    control_zone_a,
    control_zone_b,
    control_zone_c,
    control_zone_trend,
    control_zone_mixture,
    control_zone_stratification,
    control_zone_overcontrol,
)
from manufacturing.lookup_tables import (
    c4_table,
    d2_table,
    calc_B3,
    calc_B4,
    calc_D3,
    calc_D4,
)
from manufacturing.util import coerce, remove_outliers

_logger = logging.getLogger(__name__)


def ppk_plot(
    data: (List[int], List[float], pd.Series, np.ndarray),
    upper_specification_limit: (int, float),
    lower_specification_limit: (int, float),
    parameter_name: Optional[str] = None,
    threshold_percent: float = 0.001,
    is_subset: bool = False,
    figure: Optional[Figure] = None,
):
    """
    Shows the statistical distribution of the data along with CPK and limits.

    :param data: a list, pandas.Series, or numpy.ndarray representing the data set
    :param upper_specification_limit: an integer or float which represents the upper control limit, commonly called the UCL
    :param lower_specification_limit: an integer or float which represents the upper control limit, commonly called the UCL
    :param parameter_name: a string that shows up in the title
    :param threshold_percent: the threshold at which % of units above/below the number will display on the plot
    :param is_subset: False if the data represents a complete dataset, else True; determines if Ppk or Cpk are in the titles
    :param figure: an instance of ``matplotlig.axis.Axis``
    :return: ``matplotlib.figure.Figure``
    """
    plot_type = "Ppk" if not is_subset else "Cpk"

    data = coerce(data)
    data = remove_outliers(data)

    mean = data.mean()
    std = data.std()

    if figure is None:
        fig, ax = plt.subplots()
    else:
        fig = figure
        fig.clear()
        ax = fig.add_subplot(1, 1, 1)

    ax.hist(data, density=True, label="data", alpha=0.3)
    x = np.linspace(mean - 4 * std, mean + 4 * std, 100)
    pdf = stats.norm.pdf(x, mean, std)
    ax.plot(x, pdf, label="normal fit", alpha=0.7)

    bottom, top = ax.get_ylim()

    ax.axvline(mean, linestyle="--")
    ax.text(mean, top * 1.01, s=r"$\mu$", ha="center")

    ax.axvline(mean + std, alpha=0.6, linestyle="--")
    ax.text(mean + std, top * 1.01, s=r"$\sigma$", ha="center")

    ax.axvline(mean - std, alpha=0.6, linestyle="--")
    ax.text(mean - std, top * 1.01, s=r"$-\sigma$", ha="center")

    ax.axvline(mean + 2 * std, alpha=0.4, linestyle="--")
    ax.text(mean + 2 * std, top * 1.01, s=r"$2\sigma$", ha="center")

    ax.axvline(mean - 2 * std, alpha=0.4, linestyle="--")
    ax.text(mean - 2 * std, top * 1.01, s=r"-$2\sigma$", ha="center")

    ax.axvline(mean + 3 * std, alpha=0.2, linestyle="--")
    ax.text(mean + 3 * std, top * 1.01, s=r"$3\sigma$", ha="center")

    ax.axvline(mean - 3 * std, alpha=0.2, linestyle="--")
    ax.text(mean - 3 * std, top * 1.01, s=r"-$3\sigma$", ha="center")

    ax.fill_between(
        x, pdf, where=x < lower_specification_limit, facecolor="red", alpha=0.5
    )
    ax.fill_between(
        x, pdf, where=x > upper_specification_limit, facecolor="red", alpha=0.5
    )

    lower_percent = 100.0 * stats.norm.cdf(lower_specification_limit, mean, std)
    lower_percent_text = (
        f"{lower_percent:.02g}% < LSL" if lower_percent > threshold_percent else None
    )

    higher_percent = 100.0 - 100.0 * stats.norm.cdf(
        upper_specification_limit, mean, std
    )
    higher_percent_text = (
        f"{higher_percent:.02g}% > USL" if higher_percent > threshold_percent else None
    )

    left, right = ax.get_xlim()
    bottom, top = ax.get_ylim()
    cpk = calc_ppk(
        data,
        upper_specification_limit=upper_specification_limit,
        lower_specification_limit=lower_specification_limit,
    )

    lower_sigma_level = (mean - lower_specification_limit) / std
    if lower_sigma_level < 6.0:
        ax.axvline(lower_specification_limit, color="red", alpha=0.25, label="limits")
        ax.text(
            lower_specification_limit,
            top * 0.95,
            s=f"$-{lower_sigma_level:.01f}" + r"\sigma$",
            ha="center",
        )
    else:
        ax.text(left, top * 0.95, s=r"limit < $-6\sigma$", ha="left")

    upper_sigma_level = (upper_specification_limit - mean) / std
    if upper_sigma_level < 6.0:
        ax.axvline(upper_specification_limit, color="red", alpha=0.25)
        ax.text(
            upper_specification_limit,
            top * 0.95,
            s=f"${upper_sigma_level:.01f}" + r"\sigma$",
            ha="center",
        )
    else:
        ax.text(right, top * 0.95, s=r"limit > $6\sigma$", ha="right")

    strings = [f"{plot_type} = {cpk:.02g}"]

    strings.append(r"$\mu = " + f"{mean:.3g}$")
    strings.append(r"$\sigma = " + f"{std:.3g}$")

    if lower_percent_text:
        strings.append(lower_percent_text)
    if higher_percent_text:
        strings.append(higher_percent_text)

    props = dict(boxstyle="round", facecolor="white", alpha=0.75, edgecolor="grey")
    ax.text(
        right - (right - left) * 0.05,
        0.85 * top,
        "\n".join(strings),
        bbox=props,
        ha="right",
        va="top",
    )

    ax.legend(loc="lower right")

    if parameter_name is not None:
        fig.suptitle(f"{plot_type}, {parameter_name}")

    fig.tight_layout()

    return fig


def cpk_plot(
    data: (List[int], List[float], pd.Series, np.ndarray),
    upper_specification_limit: (int, float),
    lower_specification_limit: (int, float),
    parameter_name: Optional[str] = None,
    subgroup_size: int = 30,
    max_subgroups: int = 10,
    figure: Optional[Figure] = None,
) -> Figure:
    """
    Boxplot the Cpk in subgroups os size `subgroup_size`.

    :param data: a list, pandas.Series, or ``numpy.ndarray`` representing the data set
    :param upper_specification_limit: an integer or float which represents the upper specification limit, commonly called the USL
    :param lower_specification_limit: an integer or float which represents the upper specification limit, commonly called the LSL
    :param parameter_name: the name of the parameter that will be displayed on the plot
    :param subgroup_size: the number of samples to include in each subgroup
    :param max_subgroups: the maximum number of subgroups to display
    :param figure: two instances of matplotlib.axis.Axis
    :return: an instance of ``matplotlib.figure.Figure``
    """

    def chunk(seq, size):
        return (seq[pos : pos + size] for pos in range(0, len(seq), size))

    data = coerce(data)
    data = remove_outliers(data)

    # todo: offer options of historical subgrouping, such as subgroup
    #  history = 'all' or 'recent', something that
    #  allows a better historical subgrouping
    data_subgroups = []
    for i, c in enumerate(chunk(data[::-1], subgroup_size)):
        if i >= max_subgroups:
            break
        data_subgroups.append(c)

    data_subgroups = data_subgroups[::-1]

    if figure is None:
        fig, axs = plt.subplots(
            1, 2, sharey="all", gridspec_kw={"width_ratios": [4, 1]}
        )
    else:
        fig = figure
        fig.clear()
        gs = fig.add_gridspec(nrows=1, ncols=2, width_ratios=[4, 1])
        ax0 = fig.add_subplot(gs[0, 0])
        ax1 = fig.add_subplot(gs[0, 1], sharey=ax0)
        axs = [ax0, ax1]

    ax0, ax1, *_ = axs

    bp = ax1.boxplot(data, patch_artist=True)

    ax1.set_title("Ppk")
    p0, p1 = bp["medians"][0].get_xydata()
    x0, _ = p0
    x1, _ = p1
    ax1.axhline(
        upper_specification_limit, color="red", linestyle="--", zorder=-1, alpha=0.5
    )
    ax1.axhline(
        lower_specification_limit, color="red", linestyle="--", zorder=-1, alpha=0.5
    )
    ax1.set_xticks([])
    ax1.grid(color="grey", alpha=0.3)
    bp["boxes"][0].set_facecolor("lightblue")

    bps = ax0.boxplot(data_subgroups, patch_artist=True)
    ax0.set_title(f"Cpk by Subgroups, Size={subgroup_size}")
    ax0.set_xticks([])
    ax0.axhline(
        upper_specification_limit, color="red", linestyle="--", zorder=-1, alpha=0.5
    )
    ax0.axhline(
        lower_specification_limit, color="red", linestyle="--", zorder=-1, alpha=0.5
    )
    ax0.grid(color="grey", alpha=0.3)

    for box in bps["boxes"]:
        box.set_facecolor("lightblue")

    left, right = ax0.get_xlim()
    right_plus = (right - left) * 0.01 + right

    ax0.text(right_plus, upper_specification_limit, s="USL", color="red", va="center")
    ax0.text(right_plus, lower_specification_limit, s="LSL", color="red", va="center")

    cpks = []
    for i, bp_median in enumerate(bps["medians"]):
        cpk = calc_ppk(
            data_subgroups[i],
            upper_specification_limit=upper_specification_limit,
            lower_specification_limit=lower_specification_limit,
        )
        cpks.append(cpk)
    cpks = pd.Series(cpks)

    table = [f"${cpk:.02g}$" for cpk in cpks]
    ax0.table([table], rowLabels=["$Cpk$"])

    ppk = calc_ppk(
        data,
        upper_specification_limit=upper_specification_limit,
        lower_specification_limit=lower_specification_limit,
    )
    ax1.table([[f"$Ppk: {ppk:.02g}$"], [f"$Cpk_{{av}}:{cpks.mean():.02g}$"]])

    if parameter_name is not None:
        fig.suptitle(f"Cpk, {parameter_name}")

    fig.tight_layout()

    return fig


def control_plot(*args, **kwargs) -> Axis:
    """
    Depreciated - not recommended for usage.  Left for historical reasons.  Use ``control_chart`` instead.

    :param args:
    :param kwargs:
    :return:
    """
    _logger.warning(
        'control_plot function is depreciated and will be removed in a future version; use "control_chart" instead'
    )
    return control_chart_base(*args, **kwargs)


def control_chart_base(
    data: (List[int], List[float], pd.Series, np.ndarray),
    upper_control_limit: Optional[Union[int, float]] = None,
    lower_control_limit: Optional[Union[int, float]] = None,
    highlight_beyond_limits: bool = True,
    highlight_zone_a: bool = True,
    highlight_zone_b: bool = True,
    highlight_zone_c: bool = True,
    highlight_trend: bool = True,
    highlight_mixture: bool = False,
    highlight_stratification: bool = False,
    highlight_overcontrol: bool = False,
    max_points: Optional[int] = 60,
    avg_label: Optional[str] = "avg",
    show_hist: bool = True,
    ax: Optional[Axis] = None,
) -> Axis:
    """
    Create a control plot based on the input data.

    :param data: a list, pandas.Series, or numpy.ndarray representing the data set
    :param upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit; note that this is NOT the specification limit!
    :param lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limit; note that this is NOT the specification limit!
    :param highlight_beyond_limits: True if points beyond limits are to be highlighted
    :param highlight_zone_a: True if points that are zone A violations are to be highlighted
    :param highlight_zone_b: True if points that are zone B violations are to be highlighted
    :param highlight_zone_c: True if points that are zone C violations are to be highlighted
    :param highlight_trend: True if points that are trend violations are to be highlighted
    :param highlight_mixture: True if points that are mixture violations are to be highlighted
    :param highlight_stratification: True if points that are stratification violations are to be highlighted
    :param highlight_overcontrol: True if points that are overcontrol violations are to be hightlighted
    :param max_points: the maximum number of points to display ('None' to display all)
    :param avg_label: the label that is applied to the average on the plot
    :param show_hist: show a histogram to the left of the plot
    :param ax: an instance of matplotlib.axis.Axis
    :return: an instance of matplotlib.axis.Axis
    """
    data = coerce(data)

    # for purposes of gathering statistics, only use
    # data that is not considered outliers
    mean = data.mean()

    if upper_control_limit is None:
        upper_control_limit = mean + 3 * data.std()
    if lower_control_limit is None:
        lower_control_limit = mean - 3 * data.std()

    spec_range = (upper_control_limit - lower_control_limit) / 2
    spec_center = (upper_control_limit + lower_control_limit) / 2
    zone_c_upper_limit = spec_center + spec_range / 3
    zone_c_lower_limit = spec_center - spec_range / 3
    zone_b_upper_limit = spec_center + 2 * spec_range / 3
    zone_b_lower_limit = spec_center - 2 * spec_range / 3
    zone_a_upper_limit = spec_center + spec_range
    zone_a_lower_limit = spec_center - spec_range

    truncated = False
    if max_points is not None:
        _logger.info(f"data set of length {len(data)} truncated to {max_points}")
        if len(data) > max_points:
            truncated = True
        data = data[-max_points:]

    if ax is None:
        fig, ax = plt.subplots()

    ax.plot(data, marker=".")
    ax.set_title("Zone Control Chart")

    ax.axhline(spec_center, linestyle="--", color="red", alpha=0.2)

    ax.axhline(mean, linestyle="--", color="blue", alpha=0.4, zorder=-10)

    ax.set_xlim(data.index[0])
    left, right = ax.get_xlim()
    right_plus = (right - left) * 0.01 + right

    text_color = "red"
    edges = [
        zone_c_upper_limit,
        zone_c_lower_limit,
        zone_b_upper_limit,
        zone_b_lower_limit,
        spec_center,
    ]
    for edge in edges:
        ax.text(right_plus, edge, s=f"{edge:.3g}", va="center", color=text_color)

    texts = [
        {
            "y": mean,
            "s": f"{avg_label}={mean:.3g}",
            "color": "blue",
            "zorder": 100,
            "bbox": dict(
                facecolor="white", edgecolor="blue", boxstyle="round", alpha=0.8
            ),
        },
        {
            "y": upper_control_limit,
            "s": f"UCL={upper_control_limit:.3g}",
            "color": "red",
        },
        {
            "y": lower_control_limit,
            "s": f"LCL={lower_control_limit:.3g}",
            "color": "red",
        },
        {"y": (spec_center + zone_c_upper_limit) / 2, "s": "Zone C"},
        {"y": (spec_center + zone_c_lower_limit) / 2, "s": "Zone C"},
        {"y": (zone_b_upper_limit + zone_c_upper_limit) / 2, "s": "Zone B"},
        {"y": (zone_b_lower_limit + zone_c_lower_limit) / 2, "s": "Zone B"},
        {"y": (zone_a_upper_limit + zone_b_upper_limit) / 2, "s": "Zone A"},
        {"y": (zone_a_lower_limit + zone_b_lower_limit) / 2, "s": "Zone A"},
    ]
    for t in texts:
        ax.text(x=right_plus, va="center", **t)

    diameter = 30
    diameter_inc = 35
    zorder = 100
    show_legend = False
    if highlight_beyond_limits:
        beyond_limits_violations = control_beyond_limits(
            data=data,
            upper_control_limit=upper_control_limit,
            lower_control_limit=lower_control_limit,
        )
        if len(beyond_limits_violations):
            ax.scatter(
                beyond_limits_violations.index,
                beyond_limits_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                marker="o",
                label="beyond limits",
                edgecolor="red",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    if highlight_zone_a:
        zone_a_violations = control_zone_a(
            data=data,
            upper_control_limit=upper_control_limit,
            lower_control_limit=lower_control_limit,
        )
        if len(zone_a_violations):
            ax.scatter(
                zone_a_violations.index,
                zone_a_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                marker="o",
                label="zone a",
                edgecolor="orange",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    if highlight_zone_b:
        zone_b_violations = control_zone_b(
            data=data,
            upper_control_limit=upper_control_limit,
            lower_control_limit=lower_control_limit,
        )
        if len(zone_b_violations):
            ax.scatter(
                zone_b_violations.index,
                zone_b_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                label="zone b",
                edgecolor="blue",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    if highlight_zone_c:
        zone_c_violations = control_zone_c(
            data=data,
            upper_control_limit=upper_control_limit,
            lower_control_limit=lower_control_limit,
        )
        if len(zone_c_violations):
            ax.scatter(
                zone_c_violations.index,
                zone_c_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                marker="o",
                label="zone c",
                edgecolor="green",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    if highlight_trend:
        zone_trend_violations = control_zone_trend(data=data)
        if len(zone_trend_violations):
            ax.scatter(
                zone_trend_violations.index,
                zone_trend_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                marker="o",
                label="trend",
                edgecolor="purple",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    if highlight_mixture:
        zone_mixture_violations = control_zone_mixture(
            data=data,
            upper_control_limit=upper_control_limit,
            lower_control_limit=lower_control_limit,
        )
        if len(zone_mixture_violations):
            ax.scatter(
                zone_mixture_violations.index,
                zone_mixture_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                marker="o",
                label="mixture",
                edgecolor="brown",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    if highlight_stratification:
        zone_stratification_violations = control_zone_stratification(
            data=data,
            upper_control_limit=upper_control_limit,
            lower_control_limit=lower_control_limit,
        )
        if len(zone_stratification_violations):
            ax.scatter(
                zone_stratification_violations.index,
                zone_stratification_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                marker="o",
                label="stratification",
                edgecolor="orange",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    if highlight_overcontrol:
        zone_overcontrol_violations = control_zone_overcontrol(
            data=data,
            upper_control_limit=upper_control_limit,
            lower_control_limit=lower_control_limit,
        )
        if len(zone_overcontrol_violations):
            ax.scatter(
                zone_overcontrol_violations.index,
                zone_overcontrol_violations.values,
                s=diameter,
                linewidth=1,
                color="none",
                marker="o",
                label="overcontrol",
                edgecolor="magenta",
                zorder=zorder,
            )
            diameter += diameter_inc
            zorder -= 1
            show_legend = True

    q25, q75 = data.quantile(0.25), data.quantile(0.75)
    iqr = q75 - q25
    min_y = min(lower_control_limit - iqr * 0.25, mean - iqr)
    max_y = max(upper_control_limit + iqr * 0.25, mean + iqr)
    ax.set_ylim(bottom=min_y, top=max_y)

    # add background bands
    y_lower, y_upper = ax.get_ylim()
    alpha = 0.2
    ax.axhspan(y_upper, zone_a_upper_limit, color="red", alpha=alpha, zorder=-20)
    ax.axhspan(
        zone_c_upper_limit, zone_b_upper_limit, color="gray", alpha=alpha, zorder=-20
    )
    ax.axhspan(
        zone_c_lower_limit, zone_b_lower_limit, color="gray", alpha=alpha, zorder=-20
    )
    ax.axhspan(y_lower, zone_a_lower_limit, color="red", alpha=alpha, zorder=-20)

    if show_hist:
        ax_hist = ax.twiny()
        ax_hist.hist(data, density=True, orientation='horizontal',
                     zorder=-100, alpha=0.4, color='orange')
        _, xmax = ax_hist.get_xlim()
        ax_hist.set_xlim(0, xmax * 5)
        ax_hist.get_xaxis().set_visible(False)
        ax.set_zorder(1)
        ax.patch.set_visible(False)

    if truncated:
        _, y_upper = ax.get_ylim()
        x_lower, _ = ax.get_xlim()
        ax.text(x=x_lower, y=y_lower, s="...", ha="right")

    if show_legend:
        legend = ax.legend(loc="lower left")
        legend.set_zorder(200)

    fig = plt.gcf()
    fig.tight_layout()

    return ax


def x_mr_chart(
    data: (List[int], List[float], pd.Series, np.ndarray),
    parameter_name: Optional[str] = None,
    x_upper_control_limit: Optional[Union[float, int]] = None,
    x_lower_control_limit: Optional[Union[float, int]] = None,
    mr_upper_control_limit: Optional[Union[float, int]] = None,
    mr_lower_control_limit: Optional[Union[float, int]] = None,
    highlight_beyond_limits: bool = True,
    highlight_zone_a: bool = True,
    highlight_zone_b: bool = True,
    highlight_zone_c: bool = True,
    highlight_trend: bool = True,
    highlight_mixture: bool = False,
    highlight_stratification: bool = False,
    highlight_overcontrol: bool = False,
    max_points: Optional[int] = 60,
    figure: Optional[Figure] = None,
) -> Figure:
    r"""
    Create a :math:`X-mR` control plot based on the input data.

    :param data: a list, pandas.Series, or numpy.ndarray representing the data set
    :param parameter_name: a string representing the parameter name
    :param x_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the X plot; note that this is NOT the specification limit!
    :param x_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limit for the X plot; note that this is NOT the specification limit!
    :param mr_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the mR plot; note that this is NOT the specification limit!
    :param mr_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limitfor the mR plot; note that this is NOT the specification limit!
    :param highlight_beyond_limits: True if points beyond limits are to be highlighted
    :param highlight_zone_a: True if points that are zone A violations are to be highlighted
    :param highlight_zone_b: True if points that are zone B violations are to be highlighted
    :param highlight_zone_c: True if points that are zone C violations are to be highlighted
    :param highlight_trend: True if points that are trend violations are to be highlighted
    :param highlight_mixture: True if points that are mixture violations are to be highlighted
    :param highlight_stratification: True if points that are stratification violations are to be highlighted
    :param highlight_overcontrol: True if points that are overcontrol violations are to be hightlighted
    :param max_points: the maximum number of points to display ('None' to display all)
    :param figure: instance of matplotlib.figure.Figure
    :return: an instance of matplotlib.axis.Axis
    """
    data = coerce(data)
    data = remove_outliers(data)
    diff_data = abs(data.diff())

    # create an I-MR chart using a combination of control_chart and moving_range
    if figure is None:
        fig, axs = plt.subplots(2, 1, figsize=(12, 9), sharex="all")
    else:
        fig = figure
        fig.clear()
        ax0 = fig.add_subplot(211)
        ax1 = fig.add_subplot(212, sharex=ax0)
        axs = [ax0, ax1]

    params = {
        "highlight_beyond_limits": highlight_beyond_limits,
        "highlight_zone_a": highlight_zone_a,
        "highlight_zone_b": highlight_zone_b,
        "highlight_zone_c": highlight_zone_c,
        "highlight_trend": highlight_trend,
        "highlight_mixture": highlight_mixture,
        "highlight_stratification": highlight_stratification,
        "highlight_overcontrol": highlight_overcontrol,
        "max_points": max_points,
    }

    control_chart_base(
        data,
        avg_label=r"$\bar{X}$",
        upper_control_limit=x_upper_control_limit,
        lower_control_limit=x_lower_control_limit,
        ax=axs[0],
        **params,
    )

    # UCL = 1 + 3(d3 / d2) * mRbar
    #     = D4 * mRbar
    #     = 3.2665 * mRbar
    mRbar = diff_data.mean()
    if mr_upper_control_limit is not None:
        ucl = mr_upper_control_limit
    else:
        ucl = 3.2665 * mRbar

    if mr_lower_control_limit is not None:
        lcl = mr_lower_control_limit
    else:
        lcl = 0.0

    control_chart_base(
        diff_data,
        avg_label=r"$\bar{R}$",
        upper_control_limit=ucl,
        lower_control_limit=lcl,
        highlight_beyond_limits=highlight_beyond_limits,
        highlight_zone_a=False,
        highlight_zone_b=False,
        highlight_zone_c=False,
        highlight_trend=False,
        highlight_mixture=False,
        highlight_stratification=False,
        highlight_overcontrol=False,
        max_points=max_points,
        ax=axs[1],
    )

    axs[0].set_title("Individual")
    axs[1].set_title("Moving Range")

    fig_title = f"X-mR Chart"
    if parameter_name is not None:
        fig_title = f"{fig_title}, {parameter_name}"
    fig.suptitle(fig_title)

    fig.tight_layout()

    return fig


def xbar_r_chart(
    data: (List[int], List[float], pd.Series, np.ndarray),
    subgroup_size: int = 4,
    parameter_name: Optional[str] = None,
    xbar_upper_control_limit: Optional[Union[float, int]] = None,
    xbar_lower_control_limit: Optional[Union[float, int]] = None,
    r_upper_control_limit: Optional[Union[float, int]] = None,
    r_lower_control_limit: Optional[Union[float, int]] = None,
    highlight_beyond_limits: bool = True,
    highlight_zone_a: bool = True,
    highlight_zone_b: bool = True,
    highlight_zone_c: bool = True,
    highlight_trend: bool = True,
    highlight_mixture: bool = False,
    highlight_stratification: bool = False,
    highlight_overcontrol: bool = False,
    max_points: Optional[int] = 60,
    figure: Optional[Figure] = None,
) -> Figure:
    r"""
    Create a :math:`\bar{X}-R` control plot based on the input data.

    :param data: a list, pandas.Series, or numpy.ndarray representing the data set
    :param subgroup_size: an integer that determines the subgroup size
    :param parameter_name: a string representing the parameter name
    :param xbar_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the X plot; note that this is NOT the specification limit!
    :param xbar_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limit for the X plot; note that this is NOT the specification limit!
    :param r_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the R plot; note that this is NOT the specification limit!
    :param r_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limitfor the R plot; note that this is NOT the specification limit!
    :param highlight_beyond_limits: True if points beyond limits are to be highlighted
    :param highlight_zone_a: True if points that are zone A violations are to be highlighted
    :param highlight_zone_b: True if points that are zone B violations are to be highlighted
    :param highlight_zone_c: True if points that are zone C violations are to be highlighted
    :param highlight_trend: True if points that are trend violations are to be highlighted
    :param highlight_mixture: True if points that are mixture violations are to be highlighted
    :param highlight_stratification: True if points that are stratification violations are to be highlighted
    :param highlight_overcontrol: True if points that are overcontrol violations are to be hightlighted
    :param max_points: the maximum number of points to display ('None' to display all)
    :param figure: an instance of matplotlib.figure.Figure
    :return: an instance of matplotlib.figure.Figure
    """
    if subgroup_size < 2:
        raise ValueError(
            "xbar_r_chart is recommended for subgroup sizes greater than 1"
        )
    elif subgroup_size > 11:
        raise ValueError(
            "xbar_r_chart is recommended for subgroup sizes of more than 11"
        )
    data = coerce(data)
    data = remove_outliers(data)

    # determine how many arrays are in the data
    k = len(data) // subgroup_size

    # split into 'n' chunks
    x_bars = []
    ranges = []
    groups = np.array_split(data.to_numpy(), k, axis=0)
    for a in groups:
        # calculate sample average "Xbar"
        x_bars.append(a.mean())

        # calculate sample range "R"
        ranges.append(abs(max(a) - min(a)))

    n = subgroup_size

    # calculating values Xbarbar and Rbar
    x_bar_bar = sum(x_bars) / k  # average of averages (centerline of chart)
    r_bar = sum(ranges) / k
    wd = r_bar / d2_table[n]  # calculate within deviation: Wd = R_bar / d2n

    # using studentized control limits
    dev = (3 * wd) / np.sqrt(n)
    lcl_x, ucl_x = x_bar_bar - dev, x_bar_bar + dev
    lcl_r, ucl_r = calc_D3(n) * r_bar, calc_D4(n) * r_bar

    if xbar_upper_control_limit is not None:
        ucl_x = xbar_upper_control_limit
    if xbar_lower_control_limit is not None:
        lcl_x = xbar_lower_control_limit
    if r_upper_control_limit is not None:
        ucl_r = r_upper_control_limit
    if r_lower_control_limit is not None:
        lcl_r = r_lower_control_limit

    if figure is None:
        fig, axs = plt.subplots(2, 1, figsize=(12, 9), sharex="all")
    else:
        fig = figure
        fig.clear()
        ax0 = fig.add_subplot(211)
        ax1 = fig.add_subplot(212, sharex=ax0)
        axs = [ax0, ax1]

    params = {
        "highlight_beyond_limits": highlight_beyond_limits,
        "highlight_zone_a": highlight_zone_a,
        "highlight_zone_b": highlight_zone_b,
        "highlight_zone_c": highlight_zone_c,
        "highlight_trend": highlight_trend,
        "highlight_mixture": highlight_mixture,
        "highlight_stratification": highlight_stratification,
        "highlight_overcontrol": highlight_overcontrol,
        "max_points": max_points,
    }

    control_chart_base(
        x_bars,
        avg_label=r"$\bar{\bar{X}}$",
        lower_control_limit=lcl_x,
        upper_control_limit=ucl_x,
        ax=axs[0],
        **params,
    )

    control_chart_base(
        ranges,
        avg_label=r"$\bar{R}$",
        lower_control_limit=lcl_r,
        upper_control_limit=ucl_r,
        ax=axs[1],
        **params,
    )

    axs[0].set_title("Group Averages")
    axs[1].set_title("Group Ranges")

    fig_title = r"$\bar{X}-R$ Chart, n=" + f"{n}"
    if parameter_name is not None:
        fig_title = f"{fig_title}, {parameter_name}"
    fig.suptitle(fig_title)

    fig.tight_layout()

    return fig


def xbar_s_chart(
    data: (List[int], List[float], pd.Series, np.ndarray),
    subgroup_size: int = 12,
    parameter_name: Optional[str] = None,
    xbar_upper_control_limit: Optional[Union[float, int]] = None,
    xbar_lower_control_limit: Optional[Union[float, int]] = None,
    s_upper_control_limit: Optional[Union[float, int]] = None,
    s_lower_control_limit: Optional[Union[float, int]] = None,
    highlight_beyond_limits: bool = True,
    highlight_zone_a: bool = True,
    highlight_zone_b: bool = True,
    highlight_zone_c: bool = True,
    highlight_trend: bool = True,
    highlight_mixture: bool = False,
    highlight_stratification: bool = False,
    highlight_overcontrol: bool = False,
    max_points: Optional[int] = 60,
    figure: Optional[Figure] = None,
) -> Figure:
    r"""
    Create a moving :math:`\bar{X}-S` control plot based on the input data.  Recommended for datasets \
    which are to be grouped in subsets exceeding 11pcs each.

    :param data: a list, pandas.Series, or numpy.ndarray representing the data set
    :param subgroup_size: an integer that determines the subgroup size
    :param parameter_name: a string representing the parameter name
    :param xbar_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the X plot; note that this is NOT the specification limit!
    :param xbar_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limit for the X plot; note that this is NOT the specification limit!
    :param s_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the R plot; note that this is NOT the specification limit!
    :param s_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limitfor the R plot; note that this is NOT the specification limit!
    :param highlight_beyond_limits: True if points beyond limits are to be highlighted
    :param highlight_zone_a: True if points that are zone A violations are to be highlighted
    :param highlight_zone_b: True if points that are zone B violations are to be highlighted
    :param highlight_zone_c: True if points that are zone C violations are to be highlighted
    :param highlight_trend: True if points that are trend violations are to be highlighted
    :param highlight_mixture: True if points that are mixture violations are to be highlighted
    :param highlight_stratification: True if points that are stratification violations are to be highlighted
    :param highlight_overcontrol: True if points that are overcontrol violations are to be hightlighted
    :param max_points: the maximum number of points to display ('None' to display all)
    :param figure: an instance of matplotlib.figure.Figure
    :return: an instance of matplotlib.figure.Figure
    """
    if subgroup_size < 11:
        raise ValueError(
            "xbar_s_chart or x_mr_chart is recommended for "
            "subgroup sizes less than 11"
        )
    elif subgroup_size > len(c4_table):
        raise ValueError(
            f"invalid subgroup size {subgroup_size}; xbar_s_chart can currently only process subgroups"
            f"of less than {len(c4_table)}"
        )
    data = coerce(data)
    data = remove_outliers(data)

    # determine how many arrays are in the data
    k = len(data) // subgroup_size

    # split into 'n' chunks
    x_bars = []
    std_devs = []
    groups = np.array_split(data.to_numpy(), k, axis=0)
    for a in groups:
        # calculate sample average "Xbar"
        x_bars.append(a.mean())

        # calculate sample range "R"
        std_devs.append(a.std())

    n = subgroup_size
    x_bar_bar = sum(x_bars) / k  # average of averages (centerline of chart)
    s_bar = sum(std_devs) / k  # average std dev

    wd = s_bar / c4_table[n]
    dev = (3 * wd) / np.sqrt(n)
    lcl_x, ucl_x = x_bar_bar - dev, x_bar_bar + dev
    lcl_s, ucl_s = s_bar * calc_B3(n), s_bar * calc_B4(n)

    if xbar_upper_control_limit is not None:
        ucl_x = xbar_upper_control_limit
    if xbar_lower_control_limit is not None:
        lcl_x = xbar_lower_control_limit
    if s_upper_control_limit is not None:
        ucl_s = s_upper_control_limit
    if s_lower_control_limit is not None:
        lcl_s = s_lower_control_limit

    if figure is None:
        fig, axs = plt.subplots(2, 1, figsize=(12, 9), sharex="all")
    else:
        fig = figure
        fig.clear()
        ax0 = fig.add_subplot(211)
        ax1 = fig.add_subplot(212, sharex=ax0)
        axs = [ax0, ax1]

    params = {
        "highlight_beyond_limits": highlight_beyond_limits,
        "highlight_zone_a": highlight_zone_a,
        "highlight_zone_b": highlight_zone_b,
        "highlight_zone_c": highlight_zone_c,
        "highlight_trend": highlight_trend,
        "highlight_mixture": highlight_mixture,
        "highlight_stratification": highlight_stratification,
        "highlight_overcontrol": highlight_overcontrol,
        "max_points": max_points,
    }

    control_chart_base(
        x_bars,
        avg_label=r"$\bar{\bar{X}}$",
        lower_control_limit=lcl_x,
        upper_control_limit=ucl_x,
        ax=axs[0],
        **params,
    )

    control_chart_base(
        std_devs,
        avg_label=r"$\bar{S}$",
        lower_control_limit=lcl_s,
        upper_control_limit=ucl_s,
        ax=axs[1],
        **params,
    )

    axs[0].set_title("Group Averages")
    axs[1].set_title("Group Standard Deviations")

    fig_title = r"$\bar{X}-S$ Chart, n=" + f"{n}"
    if parameter_name is not None:
        fig_title = f"{fig_title}, {parameter_name}"
    fig.suptitle(fig_title)

    fig.tight_layout()

    return fig


def control_chart(
    data: (List[int], List[float], pd.Series, np.ndarray),
    parameter_name: Optional[str] = None,
    x_upper_control_limit: Optional[Union[float, int]] = None,
    x_lower_control_limit: Optional[Union[float, int]] = None,
    r_upper_control_limit: Optional[Union[float, int]] = None,
    r_lower_control_limit: Optional[Union[float, int]] = None,
    highlight_beyond_limits: bool = True,
    highlight_zone_a: bool = True,
    highlight_zone_b: bool = True,
    highlight_zone_c: bool = True,
    highlight_trend: bool = True,
    highlight_mixture: bool = False,
    highlight_stratification: bool = False,
    highlight_overcontrol: bool = False,
    max_points: Optional[int] = 60,
    figure: Optional[Figure] = None,
) -> Figure:
    """
    Automatically selects the most appropriate type of control chart, \
    based on the number of samples supplied in the data and the ``max_points``
    and returns a ``matplotlib.figure.Figure`` containing the control chart(s).

    :param data: (List[int], List[float], pd.Series, np.ndarray),
    :param parameter_name: a string representing the parameter name
    :param x_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the X plot; note that this is NOT the specification limit!
    :param x_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limit for the X plot; note that this is NOT the specification limit!
    :param r_upper_control_limit: an optional parameter which, when present, will override the internally calculated upper control limit for the R/S plot; note that this is NOT the specification limit!
    :param r_lower_control_limit: an optional parameter which, when present, will override the internally calculated lower control limitfor the R/S plot; note that this is NOT the specification limit!
    :param highlight_beyond_limits: True if points beyond limits are to be highlighted
    :param highlight_zone_a: True if points that are zone A violations are to be highlighted
    :param highlight_zone_b: True if points that are zone B violations are to be highlighted
    :param highlight_zone_c: True if points that are zone C violations are to be highlighted
    :param highlight_trend: True if points that are trend violations are to be highlighted
    :param highlight_mixture: True if points that are mixture violations are to be highlighted
    :param highlight_stratification: True if points that are stratification violations are to be highlighted
    :param highlight_overcontrol: True if points that are overcontrol violations are to be hightlighted
    :param max_points: the maximum number of points to display ('None' to display all)
    :param figure: an instance of matplotlib.figure.Figure
    :return: instance of matplotlib.figure.Figure
    """
    data = coerce(data)

    params = {
        "parameter_name": parameter_name,
        "highlight_beyond_limits": highlight_beyond_limits,
        "highlight_zone_a": highlight_zone_a,
        "highlight_zone_b": highlight_zone_b,
        "highlight_zone_c": highlight_zone_c,
        "highlight_trend": highlight_trend,
        "highlight_mixture": highlight_mixture,
        "highlight_stratification": highlight_stratification,
        "highlight_overcontrol": highlight_overcontrol,
        "max_points": max_points,
        "figure": figure,
    }

    if len(data) < max_points:
        return x_mr_chart(
            data,
            x_upper_control_limit=x_upper_control_limit,
            x_lower_control_limit=x_lower_control_limit,
            mr_upper_control_limit=r_upper_control_limit,
            mr_lower_control_limit=r_lower_control_limit,
            **params,
        )

    subgroup_size = 1 + len(data) // max_points
    if subgroup_size < 12:
        return xbar_r_chart(
            data,
            subgroup_size=subgroup_size,
            xbar_upper_control_limit=x_upper_control_limit,
            xbar_lower_control_limit=x_lower_control_limit,
            r_upper_control_limit=r_upper_control_limit,
            r_lower_control_limit=r_lower_control_limit,
            **params,
        )

    # if data is too long, then truncate
    max_subgroup_size = len(c4_table) - 1
    max_data_points = max_subgroup_size * max_points
    if len(data) > (max_data_points - 1):
        subgroup_size = max_subgroup_size
    return xbar_s_chart(
        data,
        subgroup_size=subgroup_size,
        xbar_upper_control_limit=x_upper_control_limit,
        xbar_lower_control_limit=x_lower_control_limit,
        s_upper_control_limit=r_upper_control_limit,
        s_lower_control_limit=r_lower_control_limit,
        **params,
    )


if __name__ == '__main__':
    from manufacturing.data_import import import_csv
    data = import_csv('../examples/data/example_data_with_faults.csv', columnname='value')
    data = remove_outliers(data)

    fig, ax = plt.subplots()
    control_chart_base(data=data, ax=ax)

    ax_hist = ax.twiny()
    ax_hist.hist(data, density=True, orientation='horizontal', zorder=-100, alpha=0.3, color='orange')
    _, xmax = ax_hist.get_xlim()
    ax_hist.set_xlim(0, xmax*5)
    ax_hist.get_xaxis().set_visible(False)

    fig.tight_layout()
    plt.show()



from datetime import datetime
import logging
from pathlib import Path
from os import chdir, getcwd
from os.path import splitext
from subprocess import check_call

import matplotlib.pyplot as plt
import pandas as pd

from manufacturing.analysis import normality_test, suggest_specification_limits
from manufacturing.data_import import parse_col_for_limits
from manufacturing.visual import ppk_plot, cpk_plot, control_chart

_logger = logging.getLogger(__name__)


def generate_production_report(
    input_file: (str, Path),
    output_file: Path = None,
    title: str = "Production Report",
    **kwargs,
):
    """
    Generate a report

    :param input_file: the input file to be analyzed
    :param output_file: the output file to be analyzed
    :param title: the title of the report
    :param kwargs: the keyword args to be passed into `pandas.read_csv()` or `pandas.read_excel()` methods
    :return:
    """
    _logger.info("attempting to generate report...")

    if not isinstance(input_file, Path):
        _logger.debug("coercing `str` to `Path`")
        input_file = Path(input_file)

    if input_file.name.endswith("csv"):
        _logger.debug("csv file detected")
        df = pd.read_csv(input_file, **kwargs)
    elif input_file.name.endswith("csv"):
        _logger.debug("ms excel file detected")
        df = pd.read_excel(input_file, **kwargs)
    else:
        raise ValueError("the input file extension is not currently supported")

    build_path = Path("./build")
    build_path.mkdir(parents=True, exist_ok=True)

    text = f"# {title}\n\n"
    text += f'Report generated: {datetime.now().strftime("%Y-%m-%d %H:%M")}\n\n'

    for c in df.columns:
        _logger.info(f'analyzing column "{c}"...')
        lcl, ucl = parse_col_for_limits(c)

        text += f"## Column: {c}\n\n"

        normal = normality_test(data=df[c])
        if not normal:
            text += (
                "Normality test indicates that the data is likely "
                "not normally distributed, meaning that the suggested "
                "limits are not based on a normally "
                "distributed parameter.\n\n"
            )

        if lcl is not None and ucl is not None:
            text += (
                f"Established limits:\n\n * LCL = {lcl:.02g}\n * UCL = {ucl:.02g}\n\n"
            )
        else:
            lcl, ucl = suggest_specification_limits(df[c])
            text += (
                f"Recommended limits:\n\n * LCL = {lcl:.02g}\n * UCL = {ucl:.02g}\n\n"
            )

        fig_name = c.split("(")[0].strip()

        # generate visual Ppk plot
        fig, ax = plt.subplots(1, 1)
        ppk_plot(
            df[c],
            upper_specification_limit=ucl,
            lower_specification_limit=lcl,
            figure=fig,
        )
        plot_name = build_path / f"ppk_plot_{fig_name}.png"
        fig.savefig(plot_name)
        text += f"![Ppk Plot: {fig_name}]({plot_name.name})\n\n"

        # generate zpk subgroup plot
        fig, axs = plt.subplots(1, 2, sharey=True, gridspec_kw={"width_ratios": [4, 1]})
        cpk_plot(
            df[c],
            upper_specification_limit=ucl,
            lower_specification_limit=lcl,
            figure=fig,
        )
        plot_name = build_path / f"cpk_plot_{fig_name}.png"
        fig.savefig(plot_name)
        text += f"![Cpk Plot: {fig_name}]({plot_name.name})\n\n"

        # generate zone control plot
        fig = control_chart(df[c])
        plot_name = build_path / f"zone_control_chart_{fig_name}.png"
        fig.savefig(plot_name)
        text += f"![Zone Control Plot: {fig_name}]({plot_name.name})\n\n"

        _logger.info(f'analysis of column "{c}" complete!')

    with (build_path / "report.md").open("w") as f:
        f.write(text)

    if output_file is None:
        return

    cwd = getcwd()
    chdir(build_path.absolute())

    extension = splitext(str(output_file))[1].replace(".", "")
    if extension == "pdf":
        args = ["pandoc", f"report.md", "-o", f"{title}.pdf"]
    elif extension == "html":
        args = ["pandoc", "-s", "report.md", "-o", f"{title}.html"]
    else:
        raise ValueError(f'extension "{extension}" not supported')
    _logger.info(f"executing {args}")
    check_call(args)
    chdir(cwd)

    _logger.info(
        f"report generation complete; artifact may be found at {build_path.absolute()}"
    )



"""
Contains several lookup table constants.
"""
import numpy as np

c4_table = [
    np.nan,
    np.nan,
    0.7979,
    0.8862,
    0.9213,
    0.94,
    0.9515,
    0.9594,
    0.965,
    0.9693,
    0.9727,
    0.9754,
    0.9776,
    0.9794,
    0.981,
    0.9823,
    0.9835,
    0.9845,
    0.9854,
    0.9862,
    0.9869,
    0.9876,
    0.9882,
    0.9887,
    0.9892,
    0.9896,
    0.9901,
    0.9904,
    0.9908,
    0.9911,
    0.9914,
]
d2_table = [
    np.nan,
    np.nan,
    1.1284,
    1.6926,
    2.0588,
    2.3259,
    2.5344,
    2.7044,
    2.8472,
    2.97,
    3.0775,
    3.1729,
    3.2585,
    3.336,
    3.4068,
    3.4718,
    3.532,
    3.5879,
    3.6401,
    3.689,
    3.7349,
    3.7783,
    3.8194,
    3.8583,
    3.8953,
    3.9306,
    3.9643,
    3.9965,
    4.0274,
    4.057,
    4.0855,
]
d3_table = [
    np.nan,
    np.nan,
    0.8525,
    0.8884,
    0.8798,
    0.8641,
    0.848,
    0.8332,
    0.8198,
    0.8078,
    0.7971,
    0.7873,
    0.7785,
    0.7704,
    0.763,
    0.7562,
    0.7499,
    0.7441,
    0.7386,
    0.7335,
    0.7287,
    0.7242,
    0.7199,
    0.7159,
    0.7121,
    0.7084,
    0.705,
    0.7017,
    0.6986,
    0.6955,
    0.6927,
]
d4_table = [
    np.nan,
    np.nan,
    0.9539,
    1.5878,
    1.9783,
    2.2569,
    2.4717,
    2.6455,
    2.7908,
    2.9154,
    3.0242,
    3.1205,
    3.2069,
    3.2849,
    3.3562,
    3.4217,
    3.4821,
    3.5383,
    3.5907,
    3.6398,
    3.6859,
    3.7294,
    3.7706,
    3.8096,
    3.8468,
    3.8822,
    3.9159,
    3.9482,
    3.9791,
    4.0088,
    4.0374,
]


def calc_A2(n: int) -> float:
    """
    Calculate the A2 constant for a given value `n`

    :param n: the number of samples in each group
    :return: a floating point representation of A2
    """
    try:
        d2 = d2_table[n]
    except KeyError:
        raise ValueError(f"the constant `n` must be less than {len(d2_table)}")
    if np.isnan(d2):
        raise ValueError(f"the constant `n` must be 2 or greater")
    return 3 / (d2 * np.sqrt(n))


def calc_A3(n: int) -> float:
    """
    Calculate the A3 constant for a given value `n`

    :param n: the number of samples in each group
    :return: a floating point representation of A3
    """
    try:
        c4 = c4_table[n]
    except KeyError:
        raise ValueError(f"the constant `n` must be less than {len(c4_table)}")
    if np.isnan(c4):
        raise ValueError(f"the constant `n` must be 2 or greater")
    return 3 / (c4 * np.sqrt(n))


def calc_B3(n: int) -> float:
    """
    Calculate the B3 constant for a given value `n`

    :param n: the number of samples in each group
    :return: a floating point representation of B3
    """
    try:
        c4 = c4_table[n]
    except KeyError:
        raise ValueError(f"the constant `n` must be less than {len(c4_table)}")
    if np.isnan(c4):
        raise ValueError(f"the constant `n` must be 2 or greater")

    value = 1 - (3 / c4) * np.sqrt(1 - c4**2)
    value = max(0.0, value)
    return value


def calc_B4(n: int) -> float:
    """
    Calculate the B4 constant for a given value `n`

    :param n: the number of samples in each group
    :return: a floating point representation of B4
    """
    try:
        c4 = c4_table[n]
    except KeyError:
        raise ValueError(f"the constant `n` must be less than {len(c4_table)}")
    if np.isnan(c4):
        raise ValueError(f"the constant `n` must be 2 or greater")

    return 1 + (3 / c4) * np.sqrt(1 - c4**2)


def calc_D3(n: int) -> float:
    """
    Calculate the D3 constant for a given value `n`

    :param n: the number of samples in each group
    :return: a floating point representation of D3
    """
    try:
        d3 = d3_table[n]
        d2 = d2_table[n]
    except KeyError:
        raise ValueError(f"the constant `n` must be less than {len(c4_table)}")
    if np.isnan(d3):
        raise ValueError(f"the constant `n` must be 2 or greater")

    value = 1 - 3 * d3 / d2
    value = max(0.0, value)
    return value


def calc_D4(n: int) -> float:
    """
    Calculate the D4 constant for a given value `n`

    :param n: the number of samples in each group
    :return: a floating point representation of D4
    """
    try:
        d3 = d3_table[n]
        d2 = d2_table[n]
    except KeyError:
        raise ValueError(f"the constant `n` must be less than {len(c4_table)}")
    if np.isnan(d3):
        raise ValueError(f"the constant `n` must be 2 or greater")

    return 1 + 3 * d3 / d2
